// Copyright (c) 2020, 2021 Famedly
//
// This software is released under the MIT License.
// https://opensource.org/licenses/MIT

import 'package:flutter/cupertino.dart';
import 'package:flutter/material.dart';
import 'package:provider/provider.dart';
import 'view_data.dart';

class AdaptivePageLayout extends StatefulWidget {
  final String initialRoute;
  final double columnWidth;
  final PreferredSizeWidget Function(BuildContext) topBarBuilder;
  final ViewData Function(RouteSettings) onGenerateRoute;
  final Route Function(
          Widget Function(BuildContext) builder, RouteSettings routeSettings)
      routeBuilder;
  final Color dividerColor;
  final double dividerWidth;
  final bool safeAreaOnColumnView;

  static AdaptivePageLayoutState of(BuildContext context) =>
      Provider.of<AdaptivePageLayoutState>(context, listen: false);

  /// Use this widget as the [home] of your [MaterialApp] for example and as a
  /// replacement of the [Navigator].
  ///
  /// It will use a default [Navigator] on less wide screens like mobile phones
  /// and a 2 or 3 column layout on wide screens.
  ///
  /// You have to define [onGenerateRoute] to
  /// set the existing routes of your app. Set [initialRoute] to set the initial
  /// route and [columnWidth] to change the maximum width of the left and right
  /// columns. [topBar] provides a way to display a custom topBar which you can
  /// display or hide using the [ViewData]. [routeBuilder] lets your control
  /// which RouteBuilder is used when the app is in 1-column mode.
  const AdaptivePageLayout({
    Key key,
    this.initialRoute = '/',
    this.columnWidth = 360.0,
    this.topBarBuilder,
    @required this.onGenerateRoute,
    this.routeBuilder,
    this.dividerColor = Colors.black,
    this.dividerWidth = 1.0,
    this.safeAreaOnColumnView = true,
  }) : super(key: key);
  @override
  AdaptivePageLayoutState createState() => AdaptivePageLayoutState();
}

class AdaptivePageLayoutState extends State<AdaptivePageLayout> {
  final GlobalKey<NavigatorState> _navigator = GlobalKey<NavigatorState>();
  List<ViewData> viewDataStack;
  ViewData get currentViewData => viewDataStack.last;

  @override
  void initState() {
    viewDataStack = [
      widget.onGenerateRoute(RouteSettings(name: widget.initialRoute))
    ];
    super.initState();
  }

  bool canPop() =>
      viewDataStack.length > 1 || _navigator.currentState?.canPop() == true;

  Future<void> pushNamed(String routeName, {Object arguments}) {
    viewDataStack.add(
      widget.onGenerateRoute(
        RouteSettings(name: routeName, arguments: arguments),
      ),
    );
    if (columnMode(context)) {
      setState(() {});
      return null;
    }
    return _navigator.currentState.pushNamed(routeName, arguments: arguments);
  }

  Future<void> popAndPushNamed(String routeName, {Object arguments}) {
    viewDataStack.removeLast();
    viewDataStack.add(
      widget.onGenerateRoute(
        RouteSettings(name: routeName, arguments: arguments),
      ),
    );
    if (columnMode(context)) {
      if (viewDataStack.isNotEmpty) {
        setState(() {});
      }
      return null;
    }
    return _navigator.currentState
        .popAndPushNamed(routeName, arguments: arguments);
  }

  void pop() {
    viewDataStack.removeLast();
    if (columnMode(context)) {
      if (viewDataStack.isNotEmpty) {
        setState(() {});
      }
      return;
    }
    return _navigator.currentState.pop();
  }

  void popUntilIsFirst() {
    viewDataStack.removeRange(1, viewDataStack.length);
    if (columnMode(context)) {
      setState(() {});
      return;
    }
    return _navigator.currentState.popUntil((r) => r.isFirst);
  }

  Future<T> pushNamedAndRemoveUntilIsFirst<T>(String routeName,
      {Object arguments}) {
    viewDataStack.removeRange(1, viewDataStack.length);
    viewDataStack.add(
      widget.onGenerateRoute(
        RouteSettings(name: routeName, arguments: arguments),
      ),
    );
    if (columnMode(context)) {
      setState(() {});
      return null;
    }
    return _navigator.currentState.pushNamedAndRemoveUntil(
      routeName,
      (r) => r.isFirst,
      arguments: arguments,
    );
  }

  Future<T> pushNamedAndRemoveAllOthers<T>(String routeName,
      {Object arguments}) {
    viewDataStack.clear();
    viewDataStack.add(
      widget.onGenerateRoute(
        RouteSettings(name: routeName, arguments: arguments),
      ),
    );
    if (columnMode(context)) {
      setState(() {});
      return null;
    }
    return _navigator.currentState.pushNamedAndRemoveUntil(
      routeName,
      (r) => false,
      arguments: arguments,
    );
  }

  double get _columnWidth => currentViewData?.columnWidth ?? widget.columnWidth;

  bool columnMode(BuildContext context) =>
      MediaQuery.of(context).size.width > 2.5 * _columnWidth;

  bool threeColumnMode(BuildContext context) =>
      MediaQuery.of(context).size.width > 3.5 * _columnWidth;

  void showSnackBar(SnackBar snackBar) =>
      ScaffoldMessenger.of(context).showSnackBar(snackBar);

  @override
  Widget build(BuildContext context) {
    return Provider(
      create: (_) => this,
      builder: (context, _) => WillPopScope(
        onWillPop: () {
          if (canPop()) {
            pop();
            return Future.value(false);
          }
          return Future.value(true);
        },
        child: columnMode(context)
            ? SafeArea(
                top: widget.safeAreaOnColumnView,
                left: widget.safeAreaOnColumnView,
                right: widget.safeAreaOnColumnView,
                bottom: widget.safeAreaOnColumnView,
                child: Scaffold(
                  backgroundColor: Theme.of(context).backgroundColor,
                  appBar: currentViewData.displayTopBar
                      ? widget.topBarBuilder?.call(context)
                      : null,
                  body: ScaffoldMessenger(
                    child: Row(
                      children: [
                        if (currentViewData.leftView != null ||
                            currentViewData.emptyView != null) ...{
                          Container(
                            clipBehavior: Clip.antiAlias,
                            decoration: BoxDecoration(),
                            width: _columnWidth,
                            child: currentViewData.leftView?.call(context) ??
                                currentViewData.mainView(context),
                          ),
                          Container(
                            width: widget.dividerWidth,
                            color: widget.dividerColor,
                          ),
                        },
                        if (currentViewData.rightView == null ||
                            threeColumnMode(context))
                          if (currentViewData.expandRightView)
                            Container(
                              clipBehavior: Clip.antiAlias,
                              decoration: BoxDecoration(),
                              width: _columnWidth,
                              child: currentViewData.mainView?.call(context) ??
                                  currentViewData.emptyView(context),
                            )
                          else
                            Expanded(
                              child: ClipRRect(
                                child:
                                    currentViewData.emptyView?.call(context) ??
                                        currentViewData.mainView(context),
                              ),
                            ),
                        if (threeColumnMode(context) &&
                            currentViewData.rightView == null &&
                            currentViewData.emptyView != null &&
                            currentViewData.expandRightView) ...{
                          Container(
                            width: widget.dividerWidth,
                            color: widget.dividerColor,
                          ),
                          Expanded(
                            child: ClipRRect(
                                child: currentViewData.emptyView(context)),
                          ),
                        },
                        if (currentViewData.rightView != null) ...{
                          Container(
                            width: widget.dividerWidth,
                            color: widget.dividerColor,
                          ),
                          if (!threeColumnMode(context) ||
                              currentViewData.expandRightView) ...{
                            Expanded(
                              child: ClipRRect(
                                  child: currentViewData.rightView(context)),
                            ),
                          } else
                            Container(
                              clipBehavior: Clip.antiAlias,
                              decoration: BoxDecoration(),
                              width: _columnWidth,
                              child: currentViewData.rightView(context),
                            ),
                        },
                      ],
                    ),
                  ),
                ),
              )
            : Navigator(
                key: _navigator,
                initialRoute: widget.initialRoute,
                onPopPage: (_, __) {
                  viewDataStack.removeLast();
                  return true;
                },
                onGenerateRoute: (settings) {
                  if (widget.routeBuilder != null) {
                    return widget.routeBuilder(
                      (context) =>
                          widget.onGenerateRoute(settings).focusView(context),
                      settings,
                    );
                  }
                  return MaterialPageRoute(
                    builder: (context) =>
                        widget.onGenerateRoute(settings).focusView(context),
                  );
                },
              ),
      ),
    );
  }
}
